#!/usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

import functools
import inspect
import warnings
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Sequence, Tuple, TypeVar

import numpy as np
import pandas as pd
from typing_extensions import ParamSpec

from saqc import FILTER_ALL, FILTER_NONE
from saqc.core import DictOfSeries, Flags, History
from saqc.core.translation.basescheme import TranslationScheme
from saqc.lib.docs import ParamDict, docurator
from saqc.lib.tools import isflagged, squeezeSequence, toSequence
from saqc.lib.types import ExternalFlag, OptionalNone

if TYPE_CHECKING:
    from saqc import SaQC

__all__ = [
    "register",
    "processing",
    "flagging",
]

# NOTE:
# the global SaQC function store,
# will be filled by calls to register
FUNC_MAP: Dict[str, Callable] = {}

_is_list_like = pd.api.types.is_list_like

T = TypeVar("T")
P = ParamSpec("P")


def _checkDecoratorKeywords(
    func_signature, func_name, mask, demask, squeeze, handles_target
):
    params = func_signature.parameters.keys()
    if "target" in params and not handles_target:
        raise TypeError(
            "functions defining a parameter named 'target' "
            "need to decorated with 'handles_target=True'"
        )
    for dec_arg, name in zip([mask, demask, squeeze], ["mask", "demask", "squeeze"]):
        typeerr = TypeError(
            f"type of decorator argument '{name}' must "
            f"be a list of strings, not {repr(type(dec_arg))}"
        )
        if not isinstance(dec_arg, list):
            raise typeerr
        for elem in dec_arg:
            if not isinstance(elem, str):
                raise typeerr
            if elem not in params:
                raise ValueError(
                    f"passed value {repr(elem)} in {repr(name)} is not an "
                    f"parameter in decorated function {repr(func_name)}"
                )


def _argnamesToColumns(names: list, values: dict):
    clist = []
    for name in names:
        value = values.get(name)  # eg. the value behind 'field'

        # NOTE: do not change order of the tests
        if value is None:
            pass
        elif isinstance(value, str):
            clist.append(value)
        # we ignore DataFrame, Series, DictOfSeries
        # and high order types alike
        elif hasattr(value, "columns"):
            pass
        elif _is_list_like(value) and all([isinstance(e, str) for e in value]):
            clist += value
    return pd.Index(clist)


def _warn(missing, source):
    if len(missing) == 0:
        return
    action = source + "ed"
    obj = "flags" if source == "squeeze" else "data"
    warnings.warn(
        f"Column(s) {repr(missing)} cannot not be {action} "
        f"because they are not present in {obj}. ",
        RuntimeWarning,
    )


def _getDfilter(
    func_signature: inspect.Signature,
    translation_scheme: TranslationScheme,
    kwargs: Dict[str, Any],
) -> float:
    """
    Find a default value for dfilter, either from the choosen translation scheme
    or a possibly defined method default value. Translate, if necessary.
    """
    dfilter = kwargs.get("dfilter")
    if dfilter is None or isinstance(dfilter, OptionalNone):
        # let's see, if the function has a default value
        default = func_signature.parameters.get("dfilter")
        if default is None or default.default == inspect.Signature.empty:
            default = FILTER_ALL
        else:
            default = default.default
        dfilter = max(translation_scheme.DFILTER_DEFAULT, default)
    else:
        # try to translate dfilter
        if dfilter not in {FILTER_ALL, FILTER_NONE, translation_scheme.DFILTER_DEFAULT}:
            dfilter = translation_scheme(dfilter)
    return float(dfilter)


def _squeezeFlags(old_flags, new_flags: Flags, columns: pd.Index, meta) -> Flags:
    """
    Generate flags from the temporary result-flags and the original flags.

    Parameters
    ----------
    flags : Flags
        The flags-frame, which is the result from a saqc-function

    Returns
    -------
    Flags
    """
    out = old_flags.copy()  # the old flags

    for col in columns.union(
        new_flags.columns.difference(old_flags.columns)
    ):  # account for newly added columns
        if col not in out:  # ensure existence
            out.history[col] = History(index=new_flags.history[col].index)

        old_history = out.history[col]
        new_history = new_flags.history[col]

        # We only want to add new columns, that were appended during the last
        # function call. If no such columns exist, we end up with an empty
        # new_history.
        start = len(old_history.columns)
        squeezed = new_history.squeeze(raw=True, start=start)
        out.history[col] = out.history[col].append(squeezed, meta=meta)

    return out


def _maskData(
    data: DictOfSeries, flags: Flags, columns: Sequence[str], thresh: float
) -> Tuple[DictOfSeries, DictOfSeries]:
    """
    Mask data with Nans, if the flags are worse than a threshold.
        - mask only passed `columns` (preselected by `datamask`-kw from decorator)

    Returns
    -------
    masked : DictOfSeries
        masked data, same dim as original
    mask : DictOfSeries
        dios holding iloc-data-pairs for every column in `data`
    """
    mask = DictOfSeries()

    # we use numpy here because it is faster
    for c in columns:
        col_mask = isflagged(flags[c].to_numpy(), thresh)

        if col_mask.any():
            col_data = data[c].to_numpy(dtype=np.float64, copy=True)
            mask[c] = pd.Series(col_data[col_mask], index=np.where(col_mask)[0])
            col_data[col_mask] = np.nan
            data[c] = pd.Series(col_data, index=data[c].index)

    return data, mask


def _unmaskData(
    data: DictOfSeries, mask: DictOfSeries, columns: pd.Index | None = None
) -> DictOfSeries:
    """
    Restore the masked data.

    Notes
    -----
    - Even if this returns data, it works inplace !
    - `mask` is not a boolean mask, instead it holds the original values.
      The index of mask is numeric and represent the integer location
      in the original data.
    """
    if columns is None:
        columns = data.columns  # field was in old, is in mask and is in new
    columns = mask.columns.intersection(columns)

    for c in columns:
        # ignore
        if data[c].empty or mask[c].empty:
            continue

        # get the positions of values to unmask
        candidates = mask[c]
        # if the mask was removed during the function call, don't replace
        unmask = candidates[data[c].iloc[candidates.index].isna().to_numpy()]
        if unmask.empty:
            continue
        data[c].iloc[unmask.index] = unmask

    return data


def _expandField(regex, columns, field) -> List[str]:
    """
    Expand regular expressions to concrete field names.
    """
    if regex:
        fmask = columns.str.match(field)
        return columns[fmask].tolist()
    return toSequence(field)


def _homogenizeFieldsTargets(
    multivariate,
    handles_target,
    fields,
    targets,
):
    """
    Ensure, that fields and flags are of identical length.

    Note
    ----
    We have four cases to consider:
    1. multivariate=False & handles_target=False
    2. multivariate=False & handles_target=True
    3. multivariate=True  & handles_target=False
    4. multivariate=True  & handles_target=True
    """

    if not (multivariate and handles_target):
        if len(fields) != len(targets):
            raise ValueError("expected the same number of 'field' and 'target' values")

    if multivariate:
        fields, targets = [fields], [targets]

    return fields, targets


def register(
    mask: list[str],
    demask: list[str],
    squeeze: list[str],
    multivariate: bool = False,
    handles_target: bool = False,
    docstring: dict[str, ParamDict] | None = None,
):
    """
    Generalized decorator for any saqc functions.

    Before the call of the decorated function:
    - data gets masked by flags according to `dfilter`

    After the call of the decorated function:
    - data gets demasked (original data is written back)
    - flags gets squeezed (only one history column append per call)

    Parameters
    ----------
    mask : list of string
        List of parameters of the decorated function, that specify column(s) in
        ``SaQC._data``, that are read by the function and therefore should be masked.

        The masking takes place before the call of the decorated function and
        temporary sets data to `NaN` at flagged locations. It is undone by ``demask``.
        The threshold of which data is considered to be flagged can be controlled
        via ``dfilter``, a parameter each function takes.

    demask : list of string
        List of parameters of the decorated function, that specify column(s) in
        ``SaQC._data``, that were masked (see ``mask``) and needs unmasking after the call.

        The unmasking replaces all remaining(!) ``NaN`` inserted in the masking process by
        its their original values.

    squeeze : list of string
        List of parameters of the decorated function, that specify flag column(s),
        that are written by the function.

        The squeezing combines multiple columns in the history of flags to one
        single column. This is because, multiple writes to flags, (eg. using
        ``flags[:,'a'] = 255`` twice) will result in multiple history columns,
        but should considered as a single column, because only one function call
        happened.

    multivariate : bool, default False
        If ``True``, the decorated function, processes multiple data or flag
        columns at once. Therefore the decorated function must handle a list
        of columns in the parameter ``field``.

        If ``False``, the decorated function must define ``field: str``

    handles_target : bool, default False
        If ``True``, the decorated function, handles the target parameter by
        itself. Mandatory for multivariate functions.

    docstring : dict, default None
        Allows to modify the default docstring description of the parameters ``field``,
        ``target``, ``dfilter`` and ``flag``

    """

    def outer(func: Callable[P, SaQC]) -> Callable[P, SaQC]:
        func_signature = inspect.signature(func)
        _checkDecoratorKeywords(
            func_signature, func.__name__, mask, demask, squeeze, handles_target
        )
        func = docurator(func, docstring)

        @functools.wraps(func)
        def inner(
            saqc,
            field,
            *args,
            regex: bool = False,
            flag: ExternalFlag | OptionalNone = OptionalNone(),
            **kwargs,
        ) -> "SaQC":
            # args -> kwargs
            paramnames = tuple(func_signature.parameters.keys())[
                2:
            ]  # skip (self, field)

            # check for duplicated arguments
            args_map = dict(zip(paramnames, args))
            intersection = set(args_map).intersection(set(kwargs))
            if intersection:
                raise TypeError(
                    f"SaQC.{func.__name__}() got multiple values for argument '{intersection.pop()}'"
                )
            kwargs = {**args_map, **kwargs}
            kwargs["dfilter"] = _getDfilter(func_signature, saqc._scheme, kwargs)

            # translate flag
            if not isinstance(flag, OptionalNone):
                # translation schemes might want to use a flag
                # `None` so we introduce a special class here
                kwargs["flag"] = saqc._scheme(flag)

            fields = _expandField(regex, saqc._data.columns, field)
            targets = toSequence(kwargs.pop("target", fields))

            fields, targets = _homogenizeFieldsTargets(
                multivariate, handles_target, fields, targets
            )

            out = saqc.copy(deep=True)

            # initialize target fields
            if not handles_target:
                # initialize all target variables
                for src, trg in zip(fields, targets):
                    if src != trg:
                        out = out.copyField(field=src, target=trg, overwrite=True)

            for src, trg in zip(fields, targets):
                kwargs = {**kwargs, "field": src, "target": trg}
                if not handles_target:
                    kwargs["field"] = kwargs.pop("target")

                # find columns that need masking
                # func_signature = func_signature.bind(field=field)
                columns = _argnamesToColumns(mask, kwargs)
                _warn(columns.difference(out._data.columns).to_list(), source="mask")
                columns = columns.intersection(out._data.columns)

                out._data, stored_data = _maskData(
                    data=out._data,
                    flags=out._flags,
                    columns=columns,
                    thresh=kwargs["dfilter"],
                )

                # always pass a list to multivariate functions and
                # unpack single element lists for univariate functions
                if not multivariate:
                    kwargs["field"] = squeezeSequence(kwargs["field"])

                old_flags = out._flags.copy()

                out = func(out, **kwargs)

                # find columns that need squeezing
                columns = _argnamesToColumns(squeeze, kwargs)
                _warn(
                    columns.difference(out._flags.columns).to_list(), source="squeeze"
                )
                columns = columns.intersection(out._flags.columns)

                # if the function did not want to set any flags at all,
                # we assume a processing function that altered the flags
                # in an unpredictable manner or do nothing with the flags.
                # in either case we take the returned flags as the new truth.
                if not columns.empty:
                    meta = {
                        "func": func.__name__,
                        "args": args,
                        "kwargs": kwargs,
                    }
                    out._flags = _squeezeFlags(old_flags, out._flags, columns, meta)

                # find columns that need demasking
                columns = _argnamesToColumns(demask, kwargs)
                _warn(columns.difference(out._data.columns).to_list(), source="demask")
                columns = columns.intersection(out._data.columns)

                out._data = _unmaskData(out._data, stored_data, columns=columns)
                out._validate(reason=f"call to {repr(func.__name__)}")

            return out

        FUNC_MAP[func.__name__] = inner
        return inner

    return outer


def flagging(**kwargs):
    """
    Default decorator for univariate flagging functions.

    Before the call of the decorated function:
    - `data[field]` gets masked by `flags[field]` according to `dfilter`
    After the call of the decorated function:
    - `data[field]` gets demasked (original data is written back)
    - `flags[field]` gets squeezed (only one history column append per call) if needed

    Notes
    -----
    For full control over masking, demasking and squeezing or to implement
    a multivariate function (multiple in- or outputs) use the `@register` decorator.

    See Also
    --------
        resister: generalization of of this function
    """
    if kwargs:
        raise ValueError("use '@register' to pass keywords")
    return register(mask=["field"], demask=["field"], squeeze=["field"])


def processing(**kwargs):
    """
    Default decorator for univariate processing functions.

    - no masking of data
    - no demasking of data
    - no squeezing of flags

    Notes
    -----
    For full control over masking, demasking and squeezing or to implement
    a multivariate function (multiple in- or outputs) use the `@register` decorator.

    See Also
    --------
        resister: generalization of of this function
    """
    if kwargs:
        raise ValueError("use '@register' to pass keywords")
    return register(mask=[], demask=[], squeeze=[])



# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
import pandas as pd
import pytest

from saqc.lib.tools import periodicMask

unmasked, masked = True, False


def assert_equal(result, exp):
    assert exp.index.equals(result.index)
    assert result.dtype == exp.dtype

    if result.equals(exp):
        return

    df = pd.DataFrame(
        data=np.array([result, exp]).T, index=exp.index, columns=["result", "exp"]
    )
    print("\n", df)
    assert result.equals(exp)


@pytest.mark.parametrize(
    "index",
    [
        pd.DatetimeIndex(
            [
                "1999-10",
                "1999-11",
                "1999-12",
                "2000-01",
                "2000-02",
                "2000-03",
            ]
        )
    ],
)
@pytest.mark.parametrize(
    "start, end, closed, exp",
    [
        (
            "01-01T00:00:00",
            "03-01T00:00:00",
            True,
            [unmasked, unmasked, unmasked, masked, masked, masked],
        ),
        (
            "01-01T00:00:00",
            "03-01T00:00:00",
            False,
            [unmasked, unmasked, unmasked, unmasked, masked, unmasked],
        ),
        (
            "03-01T00:00:00",
            "01-01T00:00:00",
            True,
            [masked, masked, masked, masked, unmasked, masked],
        ),
        (
            "03-01T00:00:00",
            "01-01T00:00:00",
            False,
            [masked, masked, masked, unmasked, unmasked, unmasked],
        ),
    ],
)
def test_bounds(index, start, end, closed, exp):
    exp = pd.Series(exp, index=index, dtype=bool)
    result = periodicMask(index, start, end, closed)
    assert_equal(result, exp)


@pytest.mark.parametrize(
    "index",
    [
        pd.DatetimeIndex(
            [
                "1990-01",
                "1990-02",
                "1990-03",
                "1990-04",
                "2000-01",
                "2000-02",
                "2000-03",
                "2000-04",
            ]
        )
    ],
)
@pytest.mark.parametrize(
    "start, end, closed, exp",
    [
        (
            "01-01T00:00:00",
            "03-01T00:00:00",
            True,
            [masked, masked, masked, unmasked, masked, masked, masked, unmasked],
        ),
        (
            "01-01T00:00:00",
            "03-01T00:00:00",
            False,
            [
                unmasked,
                masked,
                unmasked,
                unmasked,
                unmasked,
                masked,
                unmasked,
                unmasked,
            ],
        ),
    ],
)
def test_season(index, start, end, closed, exp):
    exp = pd.Series(exp, index=index, dtype=bool)
    result = periodicMask(index, start, end, closed)
    assert_equal(result, exp)



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Optional, Union

import numpy as np
import pandas as pd

from saqc.core import register

if TYPE_CHECKING:
    from saqc import SaQC


class TransformationMixin:
    @register(mask=["field"], demask=[], squeeze=[])
    def transform(
        self: "SaQC",
        field: str,
        func: Callable[[pd.Series | np.ndarray], pd.Series],
        freq: float | str | None = None,
        **kwargs,
    ) -> "SaQC":
        """
        Transform data by applying a custom function on data chunks of variable size. Existing flags are preserved.

        Parameters
        ----------
        func :
            Transformation function.

        freq :
            Size of the data window. The transformation is applied on each window individually

            * ``None``: Apply transformation on the entire data set at once
            * ``int`` : Apply transformation on successive data chunks of the given length. Must be grater than 0.
            * Offset String : Apply transformation on successive data chunks of the given temporal extension.
        """
        val_ser = self._data[field].copy()
        # partitioning
        if not freq:
            freq = len(val_ser)

        if isinstance(freq, str):
            grouper = pd.Grouper(freq=freq)
        else:
            grouper = pd.Series(data=np.arange(0, len(val_ser)), index=val_ser.index)
            grouper = grouper.transform(lambda x: int(np.floor(x / freq)))

        partitions = val_ser.groupby(grouper)

        for _, partition in partitions:
            if partition.empty:
                continue
            val_ser[partition.index] = func(partition)

        self._data[field] = val_ser
        return self



#! /usr/bin/env python
# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
# SPDX-License-Identifier: GPL-3.0-or-later
# -*- coding: utf-8 -*-

# isort: skip_file

"""The System for automated Quality Control package."""

from saqc.constants import BAD, DOUBTFUL, FILTER_ALL, FILTER_NONE, GOOD, UNFLAGGED
from saqc.core import Flags, DictOfSeries, SaQC
from saqc.core.translation import DmpScheme, FloatScheme, PositionalScheme, SimpleScheme
from saqc.parsing.reader import fromConfig
from saqc.version import __version__



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-


"""
Detecting breaks in data.

This module provides functions to detect and flag breaks in data, for example temporal
gaps (:py:func:`flagMissing`), jumps and drops (:py:func:`flagJumps`) or temporal
isolated values (:py:func:`flagIsolated`).
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from saqc import BAD, FILTER_ALL
from saqc.core import flagging, register
from saqc.funcs.changepoints import _getChangePoints
from saqc.lib.tools import isunflagged

if TYPE_CHECKING:
    from saqc.core.core import SaQC


class BreaksMixin:
    @register(mask=[], demask=[], squeeze=["field"])
    def flagMissing(
        self: "SaQC",
        field: str,
        flag: float = BAD,
        dfilter: float = FILTER_ALL,
        **kwargs,
    ) -> "SaQC":
        """
        Flag NaNs in data.

        By default, only NaNs are flagged, that not already have a flag.
        `dfilter` can be used to pass a flag that is used as threshold.
        Each flag worse than the threshold is replaced by the function.
        This is, because the data gets masked (with NaNs) before the
        function evaluates the NaNs.
        """

        datacol = self._data[field]
        mask = datacol.isna()

        mask = isunflagged(self._flags[field], dfilter) & mask

        self._flags[mask, field] = flag
        return self

    @flagging()
    def flagIsolated(
        self: "SaQC",
        field: str,
        gap_window: str,
        group_window: str,
        flag: float = BAD,
        **kwargs,
    ) -> "SaQC":
        """
        Find and flag temporal isolated groups of data.

        The function flags arbitrarily large groups of values, if they are surrounded by
        sufficiently large data gaps. A gap is a timespan containing either no data at all
        or NaNs only.

        Parameters
        ----------
        gap_window :
            Minimum gap size required before and after a data group to consider it
            isolated. See condition (2) and (3)

        group_window :
            Maximum size of a data chunk to consider it a candidate for an isolated group.
            Data chunks that are bigger than the ``group_window`` are ignored.
            This does not include the possible gaps surrounding it.
            See condition (1).

        Notes
        -----
        A series of values :math:`x_k,x_{k+1},...,x_{k+n}`, with associated
        timestamps :math:`t_k,t_{k+1},...,t_{k+n}`, is considered to be isolated, if:

        1. :math:`t_{k+1} - t_n <` `group_window`
        2. None of the :math:`x_j` with :math:`0 < t_k - t_j <` `gap_window`,
            is valid (preceeding gap).
        3. None of the :math:`x_j` with :math:`0 < t_j - t_(k+n) <` `gap_window`,
            is valid (succeding gap).
        """

        dat = self._data[field].dropna()
        if dat.empty:
            return self

        gap_ends = dat.rolling(gap_window).count() == 1
        gap_ends[0] = False
        gap_ends = gap_ends[gap_ends]
        gap_starts = dat[::-1].rolling(gap_window).count()[::-1] == 1
        gap_starts[-1] = False
        gap_starts = gap_starts[gap_starts]
        if gap_starts.empty:
            return self

        gap_starts = gap_starts[1:]
        gap_ends = gap_ends[:-1]
        isolated_groups = gap_starts.index - gap_ends.index < group_window
        gap_starts = gap_starts[isolated_groups]
        gap_ends = gap_ends[isolated_groups]
        to_flag = pd.Series(False, index=dat.index)
        for s, e in zip(gap_starts.index, gap_ends.index):
            # what gets flagged are the groups between the gaps, those range from
            # the end of one gap (gap_end) to the beginning of the next (gap_start)
            to_flag[e:s] = True

        to_flag = to_flag.reindex(self._data[field].index, fill_value=False)
        self._flags[to_flag.to_numpy(), field] = flag
        return self

    @flagging()
    def flagJumps(
        self: "SaQC",
        field: str,
        thresh: float,
        window: str,
        min_periods: int = 1,
        flag: float = BAD,
        dfilter: float = FILTER_ALL,
        **kwargs,
    ) -> "SaQC":
        """
        Flag jumps and drops in data.

        Flag data where the mean of its values significantly changes (, where the data "jumps" from one value level to
        another).
        The changes in value level are detected by comparing the mean for two adjacently rolling windows.
        Whenever the difference between the mean in the two windows exceeds `thresh`, the value between the windows
        is flagged a jump.

        Parameters
        ----------
        thresh :
            Threshold value by which the mean of data has to jump, to trigger flagging.

        window :
            Size of the two moving windows. This determines the number of observations used
            for calculating the mean in every window.
            The window size should be big enough to yield enough samples for a reliable mean calculation,
            but it should also not be arbitrarily big, since it also limits the density of jumps that can be detected.
            More precisely: Jumps that are not distanced to each other by more than three fourth (3/4) of the
            selected window size, will not be detected reliably.

        min_periods :
            The minimum number of observations in window required to calculate a valid
            mean value.

        Examples
        --------

        Below picture gives an abstract interpretation of the parameter interplay in case of a positive value jump,
        initialising a new mean level.

        .. figure:: /resources/images/flagJumpsPic.png

           The two adjacent windows of size `window` roll through the whole data series. Whenever the mean values in
           the two windows differ by more than `thresh`, flagging is triggered.

        Notes
        -----

        Jumps that are not distanced to each other by more than three fourth (3/4) of the
        selected window size, will not be detected reliably.
        """
        mask = _getChangePoints(
            data=self._data[field],
            stat_func=lambda x, y: np.abs(np.mean(x) - np.mean(y)),
            thresh_func=lambda x, y: thresh,
            window=window,
            min_periods=min_periods,
            result="mask",
        )

        mask = isunflagged(self._flags[field], dfilter) & mask
        self._flags[mask, field] = flag
        return self



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import pytest

import saqc
from saqc import BAD, DOUBTFUL, UNFLAGGED, SaQC
from saqc.core import DictOfSeries, initFlagsLike
from tests.common import initData
from tests.fixtures import char_dict, course_1  # noqa, todo: fix fixtures


@pytest.fixture
def data():
    return initData(cols=1, start_date="2016-01-01", end_date="2018-12-31", freq="1D")


@pytest.fixture
def field(data):
    return data.columns[0]


def test_statPass():
    data = pd.Series(0, index=pd.date_range("2000", "2001", freq="1D"), name="data")
    noise = [-1, 1] * 10
    data[100:120] = noise
    data[200:210] = noise[:10]
    data = DictOfSeries(data=data)
    flags = initFlagsLike(data)
    qc = SaQC(data, flags).flagByStatLowPass(
        "data", np.std, "20D", 0.999, "5D", 0.999, 0, flag=BAD
    )
    assert (qc.flags["data"].iloc[:100] == UNFLAGGED).all()
    assert (qc.flags["data"].iloc[100:120] == BAD).all()
    assert (qc.flags["data"].iloc[121:] == UNFLAGGED).all()


def test_flagRange(data, field):
    min, max = 10, 90
    flags = initFlagsLike(data)
    qc = SaQC(data, flags)
    qc = qc.flagRange(field, min=min, max=max, flag=BAD)
    flagged = qc.flags[field] > UNFLAGGED
    expected = (data[field] < min) | (data[field] > max)
    assert all(flagged == expected)


def test_flagSesonalRange(data, field):
    data[field].iloc[::2] = 0
    data[field].iloc[1::2] = 50
    nyears = len(data[field].index.year.unique())

    tests = [
        (
            {
                "min": 1,
                "max": 100,
                "startmonth": 7,
                "startday": 1,
                "endmonth": 8,
                "endday": 31,
            },
            31 * 2 * nyears // 2,
        ),
        (
            {
                "min": 1,
                "max": 100,
                "startmonth": 12,
                "startday": 16,
                "endmonth": 1,
                "endday": 15,
            },
            31 * nyears // 2 + 1,
        ),
    ]

    flags = initFlagsLike(data)
    qc = SaQC(data, flags)
    for test, expected in tests:
        newfield = f"{field}_masked"
        start = f"{test['startmonth']:02}-{test['startday']:02}T00:00:00"
        end = f"{test['endmonth']:02}-{test['endday']:02}T00:00:00"

        qc = qc.copyField(field, newfield)
        qc = qc.selectTime(
            newfield,
            mode="periodic",
            start=start,
            end=end,
            closed=True,
            flag=BAD,
        )
        qc = qc.flagRange(newfield, min=test["min"], max=test["max"], flag=BAD)
        qc = qc.concatFlags(
            newfield, method="match", target=field, flag=BAD, overwrite=True
        )
        qc = qc.dropField(newfield)
        flagged = qc._flags[field] > UNFLAGGED
        assert flagged.sum() == expected


def test_clearFlags(data, field):
    flags = initFlagsLike(data)
    flags[:, field] = BAD
    assert all(flags[field] == BAD)

    qc = SaQC(data, flags)
    qc = qc.clearFlags(field)
    assert all(qc._flags[field] == UNFLAGGED)


def test_forceFlags(data, field):
    flags = initFlagsLike(data)
    flags[:, field] = BAD
    assert all(flags[field] == BAD)

    qc = SaQC(data, flags).forceFlags(field, flag=DOUBTFUL)
    assert all(qc._flags[field] == DOUBTFUL)


def test_flagIsolated(data, field):
    flags = initFlagsLike(data)
    d_len = len(data[field].index)
    data[field].iloc[1:3] = np.nan
    data[field].iloc[4:5] = np.nan
    flags[data[field].index[5:6], field] = BAD
    data[field].iloc[11:13] = np.nan
    data[field].iloc[15:17] = np.nan

    #              data  flags
    # 2016-01-01   0.0   -inf
    # 2016-01-02   NaN   -inf
    # 2016-01-03   NaN   -inf
    # 2016-01-04   3.0   -inf
    # 2016-01-05   NaN   -inf
    # 2016-01-06   5.0  255.0
    # 2016-01-07   6.0   -inf
    # 2016-01-08   7.0   -inf
    #         ..    ..     ..

    qc = SaQC(data, flags).flagIsolated(
        field, group_window="1D", gap_window="2.1D", flag=BAD
    )
    assert (qc._flags[field].iloc[[3, 5]] == BAD).all()
    neg_list = [k for k in range(d_len) if k not in [3, 5]]
    assert (qc._flags[field].iloc[neg_list] == UNFLAGGED).all()

    qc = qc.flagIsolated(
        field,
        group_window="2D",
        gap_window="2.1D",
        flag=BAD,
    )
    assert (qc._flags[field].iloc[[3, 5, 13, 14]] == BAD).all()
    neg_list = [k for k in range(d_len) if k not in [3, 5, 13, 14]]
    assert (qc._flags[field].iloc[neg_list] == UNFLAGGED).all()


def test_flagManual(data, field):
    flags = initFlagsLike(data)
    dat = data[field]

    mdata = pd.Series("lala", index=dat.index)
    index_exp = mdata.iloc[[10, 33, 200, 500]].index
    mdata.iloc[[101, 133, 220, 506]] = "b"
    mdata.loc[index_exp] = "a"
    shrinked = mdata.loc[index_exp.union(mdata.iloc[[1, 2, 3, 4, 600, 601]].index)]

    kwargs_list = [
        dict(mdata=mdata, mflag="a", method="plain", mformat="mflag", flag=BAD),
        dict(mdata=mdata, mflag="a", method="ontime", mformat="mflag", flag=BAD),
        dict(mdata=shrinked, mflag="a", method="ontime", mformat="mflag", flag=BAD),
    ]

    for kw in kwargs_list:
        qc = SaQC(data, flags).flagManual(field, **kw)
        isflagged = qc._flags[field] > UNFLAGGED
        assert isflagged[isflagged].index.equals(index_exp)

    # flag not exist in mdata
    qc = SaQC(data, flags).flagManual(
        field,
        mdata=mdata,
        mflag="i do not exist",
        method="ontime",
        mformat="mflag",
        flag=BAD,
    )
    isflagged = qc._flags[field] > UNFLAGGED
    assert isflagged[isflagged].index.equals(pd.DatetimeIndex([]))

    # check closure methods
    index = pd.date_range(start="2016-01-01", end="2018-12-31", periods=11)
    mdata = pd.Series(0, index=index)
    mdata.loc[index[[1, 5, 6, 7, 9, 10]]] = 1
    # >>> mdata
    # 2016-01-01 00:00:00    0
    # 2016-04-19 12:00:00    1
    # 2016-08-07 00:00:00    0
    # 2016-11-24 12:00:00    0
    # 2017-03-14 00:00:00    0
    # 2017-07-01 12:00:00    1
    # 2017-10-19 00:00:00    1
    # 2018-02-05 12:00:00    1
    # 2018-05-26 00:00:00    0
    # 2018-09-12 12:00:00    1
    # 2018-12-31 00:00:00    1
    # dtype: int64
    m_index = mdata.index
    flag_intervals = [
        (m_index[1], m_index[2]),
        (m_index[5], m_index[8]),
        (m_index[9], dat.index.shift(freq="1h")[-1]),
    ]
    bound_drops = {"right-open": [1], "left-open": [0], "closed": []}
    for method in ["right-open", "left-open", "closed"]:
        qc = qc.flagManual(
            field,
            mdata=mdata,
            mflag=1,
            method=method,
            mformat="mflag",
            flag=BAD,
        )
        isflagged = qc._flags[field] > UNFLAGGED
        for flag_i in flag_intervals:
            f_i = isflagged[slice(flag_i[0], flag_i[-1])].index
            check_i = f_i.drop(
                [flag_i[k] for k in bound_drops[method]], errors="ignore"
            )
            assert isflagged[check_i].all()
            unflagged = isflagged[f_i.difference(check_i)]
            if not unflagged.empty:
                assert ~unflagged.all()


@pytest.mark.parametrize("dat", [pytest.lazy_fixture("course_1")])
def test_flagDriftFromNorm(dat):
    data = dat(periods=200, peak_level=5, name="field1")[0]
    data["field2"] = dat(periods=200, peak_level=10, name="field2")[0]["field2"]
    data["field3"] = dat(periods=200, peak_level=100, name="field3")[0]["field3"]

    fields = ["field1", "field2", "field3"]

    flags = initFlagsLike(data)
    qc = SaQC(data, flags).flagDriftFromNorm(
        field=fields,
        freq="200min",
        spread=5,
        flag=BAD,
    )
    assert all(qc._flags["field3"] > UNFLAGGED)


@pytest.mark.parametrize("dat", [pytest.lazy_fixture("course_1")])
def test_flagDriftFromReference(dat):
    data = dat(periods=200, peak_level=5, name="field1")[0]
    data["field2"] = dat(periods=200, peak_level=10, name="field2")[0]["field2"]
    data["field3"] = dat(periods=200, peak_level=100, name="field3")[0]["field3"]

    fields = ["field1", "field2", "field3"]

    flags = initFlagsLike(data)

    qc = SaQC(data, flags).flagDriftFromReference(
        field=fields,
        reference="field1",
        freq="3D",
        thresh=20,
        flag=BAD,
    )
    assert all(qc._flags["field3"] > UNFLAGGED)


def test_transferFlags():
    data = pd.DataFrame({"a": [1, 2], "b": [1, 2], "c": [1, 2]})
    qc = saqc.SaQC(data)
    qc = qc.flagRange("a", max=1.5)
    with pytest.deprecated_call():
        qc = qc.transferFlags(["a", "a"], ["b", "c"])  # noqa
        assert np.all(qc.flags["b"].values == np.array([UNFLAGGED, BAD]))
        assert np.all(qc.flags["c"].values == np.array([UNFLAGGED, BAD]))


def test_flagJumps():
    data = pd.DataFrame(
        {"a": [1, 1, 1, 1, 1, 6, 6, 6, 6, 6]},
        index=pd.date_range(start="2020-01-01", periods=10, freq="D"),
    )
    qc = SaQC(data=data)
    qc = qc.flagJumps(field="a", thresh=1, window="2D")
    assert qc.flags["a"][5] == BAD
    assert np.all(qc.flags["a"].values[:5] == UNFLAGGED) & np.all(
        qc.flags["a"].values[6:] == UNFLAGGED
    )



#! /usr/bin/env python

# SPDX-FileCopyrightText: 2021 Helmholtz-Zentrum für Umweltforschung GmbH - UFZ
#
# SPDX-License-Identifier: GPL-3.0-or-later

# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import scipy.stats as st

import saqc.lib.ts_operators as ts_ops
from saqc import BAD, DOUBTFUL, FILTER_ALL, FILTER_NONE, GOOD, UNFLAGGED


def clip(series, lower=None, upper=None):
    return series.clip(lower=lower, upper=upper)


def zscore(obj):
    return st.zscore(obj, nan_policy="omit")


def cv(series: pd.Series) -> pd.Series:
    """
    calculates the coefficient of variation on a min-max scaled time series
    """
    series_ = (series - series.min()) / (series.max() - series.min())
    return series_.std() / series_.mean()


ENVIRONMENT = {
    # Infinity constant
    "inf": np.inf,
    "INF": np.inf,
    # Not a number constant.
    "NAN": np.nan,
    "nan": np.nan,
    # Absolute value function.
    "abs": np.abs,
    # Maximum value function. Ignores NaN.
    "max": np.nanmax,
    # Minimum Value function. Ignores NaN.
    "min": np.nanmin,
    # Mean value function. Ignores NaN.
    "mean": np.nanmean,
    # Summation. Ignores NaN.
    "sum": np.nansum,
    # Standard deviation. Ignores NaN.
    "len": len,
    # exponential function.
    "exp": np.exp,
    # Logarithm.
    "log": np.log,
    # Logarithm, returning NaN for zero input, instead of -inf.
    "nanLog": ts_ops.zeroLog,
    # Standard deviation. Ignores NaN.
    "std": np.nanstd,
    # Variance. Ignores NaN.
    "var": np.nanvar,
    # Coefficient of variation.
    "cv": cv,
    # Median. Ignores NaN.
    "median": np.nanmedian,
    # Count Number of values. Ignores NaNs.
    "count": ts_ops.count,
    # Identity.
    "id": ts_ops.identity,
    # Returns a series` diff.
    "diff": ts_ops.difference,
    # Scales data to [0,1] interval.
    "scale": ts_ops.normScale,
    # Standardize with standard deviation.
    "zScore": zscore,
    # Standardize with median and MAD.
    "madScore": ts_ops.standardizeByMedian,
    # Standardize with median and inter quantile range.
    "iqsScore": ts_ops.standardizeByIQR,
    "clip": clip,
    "GOOD": GOOD,
    "BAD": BAD,
    "UNFLAGGED": UNFLAGGED,
    "DOUBTFUL": DOUBTFUL,
    "FILTER_ALL": FILTER_ALL,
    "FILTER_NONE": FILTER_NONE,
}
